import numpy as np
import h5py


def iwmesh(iw_list, beta):
    """Defines imaginary frequency grid from imaginary time grid.
    """
    return 1j * (2 * iw_list + 1) * np.pi / beta


def get_tau_mesh(beta, ncheb):
    """Generate tau grid for a given beta in the chebishev basis.
    """
    nts = ncheb + 2
    tau_mesh = np.zeros(nts)
    tau_mesh[0], tau_mesh[-1] = 0, beta
    # Define tau mesh
    for k in range(1, nts - 1):
        xk = np.cos(np.pi * (k - 0.5) / ncheb)
        tau_mesh[nts - k - 1] = (xk + 1) * beta / 2
    return tau_mesh


def get_Cheby(ncheb, tau_mesh):
    nts = ncheb + 2
    beta = tau_mesh[-1]
    # Define Chebyshev polynomials
    _Ttc = np.zeros((nts, ncheb))
    for it in range(nts):
        x = 2.0 * tau_mesh[it] / beta - 1.0
        _Ttc[it, 0] = 1.0
        _Ttc[it, 1] = x
        for ic in range(2, ncheb):
            _Ttc[it, ic] = 2.0 * x * _Ttc[it, ic - 1] - _Ttc[it, ic - 2]
    return _Ttc


###############
# Input    - Fermionic object in tau domain (nts, nk, nao, nao)
#
# Output - Fermionic object in Matsubara domain (nk, nao, nao)
###############
# TODO make beta as a parameter
# TODO make TNC.h5 path as a parameter and should be initialized at mb class.
# TODO make it compatible with local object (i.e no k-dependency)
def sigma_w_uniform(Sigma_tau, nw):
    beta = 100
    nk, nao = np.shape(Sigma_tau)[1:3]
    nts, ncheb = np.shape(Sigma_tau)[0], np.shape(Sigma_tau)[0] - 2
    # Define tau mesh on nodes
    tau_mesh = get_tau_mesh(beta, ncheb)
    # Compute Chebyshev on nodes
    _Ttc = get_Cheby(ncheb, tau_mesh)

    normal = 1.0/ncheb
    _Tct = np.zeros((ncheb, nts))
    for ic in range(ncheb):
        if ic == 0:
            factor = 1.0
        else:
            factor = 2.0
        for it in range(1, nts-1):
            _Tct[ic, it] = _Ttc[it, ic] * factor * normal

    _Tnc = np.zeros((nw, ncheb), dtype=complex)
    Tnc = h5py.File('/Users/CanonYeh/Projects/chebyshev_input/TNC.h5')
    for ic in range(ncheb):
        re = Tnc["TNC_" + str(ic) + "_r"][()]
        im = Tnc["TNC_" + str(ic) + "_i"][()]
        for iw in range(nw):
            _Tnc[iw, ic] = complex(re[iw], im[iw]) * beta / 2.0
    Tnc.close()

    Sigma_tau = Sigma_tau.reshape(nts, nk*nao*nao)
    # Selfenergy in Chebyshev representation
    Sigma_c = np.einsum('ij,jk -> ik', _Tct, Sigma_tau)
    # Selfenergy in Matsubara axes
    Sigma_w = np.einsum('ij,jk->ik', _Tnc, Sigma_c)
    Sigma_w = Sigma_w.reshape(nw, nk, nao, nao)
    return Sigma_w


###############
# Input    - Fermionic object in tau domain (nts, nk, nao, nao)
#
# Output - Fermionic object in Matsubara domain at n = 0 (nk, nao, nao)
# TODO make beta as a parameter
###############
def sigma_zero(Sigma):
    # Setup
    beta = 100
    nk, nao = np.shape(Sigma)[1:3]
    nts, ncheb = np.shape(Sigma)[0], np.shape(Sigma)[0] - 2
    tau_mesh = get_tau_mesh(beta, ncheb)
    # Define Chebyshev polynomials
    _Ttc = get_Cheby(ncheb, tau_mesh)

    normal = 1.0/ncheb
    _Tct = np.zeros((ncheb, nts))
    for ic in range(ncheb):
        if ic == 0:
            factor = 1.0
        else:
            factor = 2.0
        for it in range(1, nts-1):
            _Tct[ic, it] = _Ttc[it, ic] * factor * normal

    _T_0l = np.zeros(ncheb, dtype=complex)
    Tnl = h5py.File('/Users/CanonYeh/Projects/chebyshev_input/TNC.h5')
    for ic in range(ncheb):
        re = Tnl["TNC_"+str(ic)+"_r"][()]
        im = Tnl["TNC_"+str(ic)+"_i"][()]
        _T_0l[ic] = complex(re[0], im[0]) * beta / 2.0
    Tnl.close()

    Sigma = Sigma.reshape(nts, nk*nao*nao)
    # Selfenergy in Chebyshev representation
    Sigma_c = np.dot(_Tct, Sigma)
    # Sigma_c = np.einsum('ij,jklm -> iklm', _Tct, Sigma)
    # Selfenergy in zero frequency
    Sigma_w0 = np.dot(_T_0l, Sigma_c)
    # Sigma_w0 = np.einsum('i,ijkl->jkl', _T_0l, Sigma_c)

    Sigma_w0 = Sigma_w0.reshape(nk, nao, nao)
    return Sigma_w0



import numpy as np
import scipy.linalg as LA
from . import orth

#################
# Input - Fock matrix. Dim = (ns, nk, nao, nao)
# Input - Unrestricted or not
# Input - return quasi-particle basis or not
#
# Output - Quasi-particle states
#################


def compute_mo(F, S, eigh_solver=LA.eigh, thr=1e-7):
    '''
    Solve the generalized eigen problem: FC = SCE

    :param F: Fock matrix, dim = (ns, nk, nao, nao)
    :param S: Overlap matrix, dim = (ns, nk, nao, nao)
    :param eigh_solver: eigenvalue problem solver
    :param thr: lowest eigenvalues of S.
    Only used in canonical orthogonalization.

    :return:
    eig_sk: molecular energies
    mo_coeff_k: molecular orbital coefficient in AO basis
    '''
    ns, nk, nao = F.shape[0:3]
    eiv_sk = np.zeros((ns, nk, nao))
    mo_coeff_sk = np.zeros((ns, nk, nao, nao), dtype=F.dtype)
    # eiv_sk = []
    # mo_coeff_sk = []
    if S is None:
        S = np.array([[np.eye(nao)]*nk]*ns)
    for ss in range(ns):
        for k in range(nk):
            eiv, mo = eigh_solver(F[ss, k], S[ss, k], thr)
            # Re-order
            idx = np.argmax(abs(mo.real), axis=0)
            mo[:, mo[idx, np.arange(len(eiv))].real < 0] *= -1
            nbands = eiv.shape[0]
            eiv_sk[ss, k, :nbands] = eiv
            mo_coeff_sk[ss, k, :, :nbands] = mo
            # eiv_sk.append(eiv)
            # mo_coeff_sk.append(mo)

    # eig_sk = np.asarray(eiv_sk).reshape(ns, nk, nao)
    # mo_coeff_sk = np.asarray(mo_coeff_sk).reshape(ns, nk, nao, nao)

    return eiv_sk, mo_coeff_sk


def compute_no(dm, S=None):
    """Compute natural orbitals by diagonalizing density matrix
    :return:
    """
    ns, ink = dm.shape[0], dm.shape[1]
    dm_orth = orth.sao_orth(dm, S, 'g') if S is not None else dm.copy()
    occ = np.zeros(np.shape(dm)[:-1])
    no_coeff = np.zeros(np.shape(dm), dtype=complex)
    for ss in range(ns):
        for ik in range(ink):
            occ[ss, ik], no_coeff[ss, ik] = np.linalg.eigh(dm_orth[ss, ik])

    occ, no_coeff = occ[:, :, ::-1], no_coeff[:, :, :, ::-1]
    return occ, no_coeff



import numpy
import h5py
import logging
import scipy.linalg as LA

from . import integral_utils as int_utils
from . import common_utils as comm

from pyscf import lib
from pyscf.pbc.lib.kpts_helper import is_zero, member, unique
from pyscf.pbc.df import df, ft_ao, incore, aft
from pyscf.pbc.df.df_jk import zdotCN
import importlib.util as iu
if iu.find_spec('pyscf.pbc.df.gdf_builder') is not None :
    import pyscf.pbc.df.gdf_builder as gdf_builder

'''
Gaussian density fitting with the overlap metric
'''

def make_kptij_lst(kpts, kpts_band = None):
    if kpts_band is None:
        kband_uniq = numpy.zeros((0,3))
    else:
        kband_uniq = [k for k in kpts_band if len(member(k, kpts))==0]
    kptij_lst = [(ki, kpts[j]) for i, ki in enumerate(kpts) for j in range(i+1)]
    kptij_lst.extend([(ki, kj) for ki in kband_uniq for kj in kpts])
    kptij_lst.extend([(ki, ki) for ki in kband_uniq])
    kptij_lst = numpy.asarray(kptij_lst)

    return kptij_lst

def compute_j2c_sqrt(uniq_kptji_id, j2c, linear_dep_threshold=1e-9):
    try:
        # Cholesky decompose j2c
        j2c_sqrt = LA.cholesky(j2c, lower=False)
        j2ctag = 'CD'
    except LA.LinAlgError:
        w, v = LA.eigh(j2c)
        logging.info(f"DF metric linear dependency for kpt {uniq_kptji_id}")
        logging.info("cond = {}, drop {} bfns for lin_dep_threshold = {}".format(w[-1]/w[0], numpy.count_nonzero(w<linear_dep_threshold), linear_dep_threshold))
        v1 = v[:,w>linear_dep_threshold].conj().T
        v1 *= numpy.sqrt(w[w>linear_dep_threshold]).reshape(-1,1)
        j2c_sqrt = v1
        w = v = None
        j2ctag = 'eig'
  
    return j2c_sqrt, j2ctag

def _make_j2c_rsgdf(mydf, cell, auxcell, uniq_kpts, exx=False):
    from pyscf.pbc.df import rsdf_helper
    from pyscf.pbc.df.rsdf import get_aux_chg
    nao, naux = cell.nao_nr(), auxcell.nao_nr()
    # get charge of auxbasis
    if cell.dimension == 3:
      qaux = get_aux_chg(auxcell)
    else:
      qaux = numpy.zeros(auxcell.nao_nr())
  
    omega_j2c = abs(mydf.omega_j2c)
    j2c = rsdf_helper.intor_j2c(auxcell, omega_j2c, kpts=uniq_kpts)
  
    # Add (1) short-range G=0 (i.e., charge) part and (2) long-range part
    qaux2 = None
    g0_j2c = numpy.pi/omega_j2c**2./cell.vol
    mesh_j2c = mydf.mesh_j2c
    Gv, Gvbase, kws = cell.get_Gv_weights(mesh_j2c)
    b = cell.reciprocal_vectors()
    gxyz = lib.cartesian_prod([numpy.arange(len(x)) for x in Gvbase])
    ngrids = gxyz.shape[0]
  
    max_memory = max(2000, mydf.max_memory - lib.current_memory()[0])
    blksize = max(2048, int(max_memory*.5e6/16/auxcell.nao_nr()))
  
    for k, kpt in enumerate(uniq_kpts):
        # short-range charge part
        if is_zero(kpt) and cell.dimension == 3:
            if qaux2 is None:
                qaux2 = numpy.outer(qaux,qaux)
            j2c[k] -= qaux2 * g0_j2c
        # long-range part via aft
        coulG_lr = mydf.weighted_coulG(kpt, False, mesh_j2c, omega_j2c)
        for p0, p1 in lib.prange(0, ngrids, blksize):
            aoaux = ft_ao.ft_ao(auxcell, Gv[p0:p1], None, b, gxyz[p0:p1],
                                Gvbase, kpt).T
            LkR = numpy.asarray(aoaux.real, order='C')
            LkI = numpy.asarray(aoaux.imag, order='C')
            aoaux = None
  
            if is_zero(kpt):  # kpti == kptj
                j2c[k] += lib.ddot(LkR*coulG_lr[p0:p1], LkR.T)
                j2c[k] += lib.ddot(LkI*coulG_lr[p0:p1], LkI.T)
            else:
                j2cR, j2cI = df.df_jk.zdotCN(LkR*coulG_lr[p0:p1],
                                             LkI*coulG_lr[p0:p1], LkR.T, LkI.T)
                j2c[k] += j2cR + j2cI * 1j
  
            LkR = LkI = None
  
    coulG_lr = None

    return j2c

# make_j2c is updated to be consistent with PySCF 2.0.1 
# where the range of lattice Ls vectors are improved. 
# This version should be compatible with both PySCF 1.7 and 2.0
def _make_j2c_gdf(mydf, cell, auxcell, uniq_kpts, exx=False):
    import importlib.util as iu
    if iu.find_spec('pyscf.pbc.df.gdf_builder') is not None :
        fused_cell, fuse = gdf_builder.fuse_auxcell(auxcell, aft.estimate_eta(cell, cell.precision))
    else : 
        fused_cell, fuse = df.fuse_auxcell(mydf, auxcell)
  
    nao = cell.nao_nr()
    naux = auxcell.nao_nr()
    mesh = mydf.mesh
    Gv, Gvbase, kws = cell.get_Gv_weights(mesh)
    b = cell.reciprocal_vectors()
    gxyz = lib.cartesian_prod([numpy.arange(len(x)) for x in Gvbase])
    ngrids = gxyz.shape[0]
    
    # Real space integrals within an given amount of repeated unit cell images 
    j2c = fused_cell.pbc_intor('int2c2e', hermi=0, kpts=uniq_kpts) 
  
    max_memory = max(2000, mydf.max_memory - lib.current_memory()[0])
    blksize = max(2048, int(max_memory*.5e6/16/fused_cell.nao_nr()))
    # Long-range contributions of the j2c integral are evaluated analytically in the reciprocal basis, 
    # i.e. the plane-wave basis. The G=0 contribution is excluded unless specified.
    for k, kpt in enumerate(uniq_kpts):
        if exx == 'ewald': 
            print("Enable Ewald correction!")
            coulG = mydf.weighted_coulG(kpt, exx, mesh)
        else:
            coulG = mydf.weighted_coulG(kpt, False, mesh)
        for p0, p1 in lib.prange(0, ngrids, blksize):
            aoaux = ft_ao.ft_ao(fused_cell, Gv[p0:p1], None, b, gxyz[p0:p1], Gvbase, kpt).T
            LkR = numpy.asarray(aoaux.real, order='C')
            LkI = numpy.asarray(aoaux.imag, order='C')
            aoaux = None
      
            if is_zero(kpt):  # kpti == kptj
                j2c_p  = lib.ddot(LkR[naux:]*coulG[p0:p1], LkR.T)
                j2c_p += lib.ddot(LkI[naux:]*coulG[p0:p1], LkI.T)
            else:
                j2cR, j2cI = zdotCN(LkR[naux:]*coulG[p0:p1],
                                  LkI[naux:]*coulG[p0:p1], LkR.T, LkI.T)
                j2c_p = j2cR + j2cI * 1j
            j2c[k][naux:] -= j2c_p
            j2c[k][:naux,naux:] -= j2c_p[:,:naux].conj().T
            j2c_p = LkR = LkI = None
        # Symmetrizing the matrix is not must if the integrals converged.
        # Since symmetry cannot be enforced in the pbc_intor('int2c2e'),
        # the aggregated j2c here may have error in hermitian if the range of
        # lattice sum is not big enough.
        j2c[k] = (j2c[k] + j2c[k].conj().T) * .5
        j2c[k] = fuse(fuse(j2c[k]).T).T

    return j2c

def sqrt_j2c(mydf, j2c):
    j2ctags = []
    for iq in range(len(j2c)):
      # j2c_sqrt: (naux_effective, naux). naux_effective <= naux due to linear dependency
      j2c[iq], tag = compute_j2c_sqrt(iq, j2c[iq], mydf.linear_dep_threshold)
      j2ctags.append(tag)

    return j2c, j2ctags

def make_j2c_sqrt(mydf, cell):
    make_j2c = _make_j2c_rsgdf
  
    kmesh_scaled = cell.get_scaled_kpts(mydf.kpts)
    kptij_lst = make_kptij_lst(kmesh_scaled)
    nkptij = len(kptij_lst)
    kptis = kptij_lst[:,0]
    kptjs = kptij_lst[:,1]
    kpt_ji = kptjs - kptis
    # Fold kpt_ji back to [-0.5, 0.5] BZ notation
    kpt_ji = comm.fold_back_to_1stBZ(kpt_ji)
    # uniq_qs: unique q-points, # uniq_inverse: kptij -> q
    uniq_qs, uniq_index, uniq_inverse = unique(kpt_ji)
    q_ir_list, q_index, _, q_conj_list = comm.inversion_sym(uniq_qs)
    logging.info("number of k-pairs: {}".format(len(kptij_lst)))
    logging.info("number of uniq q-pionts: {}".format(len(uniq_qs)))
    logging.info("number of uniq q-points w/ the inversion symmetry: {}".format(len(q_ir_list)))
  
    kptij_lst = cell.get_abs_kpts(kptij_lst)
    uniq_qs = cell.get_abs_kpts(uniq_qs)
  
    # j2c: [q_ir_list][naux, naux]
    j2c = make_j2c(mydf, cell, mydf.auxcell, uniq_qs[q_ir_list])
    j2c, j2ctags = sqrt_j2c(mydf, j2c)
  
    return j2c, uniq_qs


def make_j3c_outcore(mydf, cell, basename = 'df_int', rsgdf=False, j2c_sqrt=True, exx=False):
    '''
    The outcore version of make_j3c
    '''
    import os
    if not rsgdf:
        make_j2c = _make_j2c_gdf
    else:
        make_j2c = _make_j2c_rsgdf
    nao, naux = cell.nao_nr(), mydf.auxcell.nao_nr()
  
    a_lattice = cell.lattice_vectors() / (2*numpy.pi)
    kmesh_scaled = cell.get_scaled_kpts(mydf.kpts)
    kptij_lst = make_kptij_lst(kmesh_scaled)
    kptij_idx = [(i, j) for i in range(mydf.kpts.shape[0]) for j in range(i+1)]
    kptij_idx = numpy.asarray(kptij_idx)
    nkptij = len(kptij_lst)
    kptis = kptij_lst[:,0]
    kptjs = kptij_lst[:,1]
    kpt_ji = kptjs - kptis
    kpt_ji = comm.fold_back_to_1stBZ(kpt_ji)
    # uniq_qs: unique q-points, # uniq_inverse: kptij -> q
    uniq_qs, uniq_index, uniq_inverse = unique(kpt_ji)
    q_ir_list, q_index, _, q_conj_list = comm.inversion_sym(uniq_qs)
    # Reduced kpair list
    kptis, kptjs = cell.get_abs_kpts(kptis), cell.get_abs_kpts(kptjs)
    kij_conj, kij_trans = int_utils.kpair_reduced_lists(kptis, kptjs, kptij_idx, mydf.kpts, a_lattice)
    kpair_irre_list = numpy.argwhere(kij_conj == kij_trans)[:,0]
    num_kpair_stored = len(kpair_irre_list)
    print("number of k-pairs: {}".format(len(kptij_lst)))
    print("number of reduced k-pairs: ", num_kpair_stored)
    print("number of uniq q-pionts: ", len(uniq_qs))
    print("number of uniq q-points w/ the inversion symmetry: ", len(q_ir_list))
  
    kptij_lst = cell.get_abs_kpts(kptij_lst)
    uniq_qs = cell.get_abs_kpts(uniq_qs)
  
    single_rho_size = nao**2 * naux * 16
    full_rho_size   = (num_kpair_stored * single_rho_size)
    chunk_size = int_utils.compute_partitioning(full_rho_size, num_kpair_stored)
    num_chunks = int(numpy.ceil(num_kpair_stored / chunk_size))
    print("Chunk size: {}, number of chunks: {}".format(chunk_size, num_chunks))
  
    # j2c: [q_ir_list][naux, naux]
    print("Computing j2c...")
    j2c = make_j2c(mydf, cell, mydf.auxcell, uniq_qs[q_ir_list], exx)
    j2c, j2ctags = sqrt_j2c(mydf, j2c)
    print("Done!")
  
    # ovlp_2c1e: [q_ir_list][naux, naux]
    # Since the memory requirement is not high, this will be stored on the fly 
    print("Computing ovlp_2c1e...")
    ovlp_2c1e_aux = mydf.auxcell.pbc_intor('int1e_ovlp', hermi=1, kpts=uniq_qs[q_ir_list])
    print("Done!")
  
    # loop over chunks
    filename = basename + "/meta.h5"
    os.system("sync")
    if os.path.exists(basename):
        os.system("rm -r " + basename)  
        os.system("sync")
    os.system("mkdir -p " + basename)
  
    ovlp_3c1e = numpy.zeros((chunk_size, nao*nao, naux), dtype=complex)
    c0 = 0
    for i in range(num_chunks):
        c1 = c0 + chunk_size
        if c1 >= num_kpair_stored:
            c1 = num_kpair_stored
        local_kpairs_num = c1 - c0
        kij_idx_local = kpair_irre_list[c0:c1]
        ovlp_3c1e[:local_kpairs_num] = incore.aux_e2(cell, mydf.auxcell, intor='int3c1e', kptij_lst=kptij_lst[kij_idx_local])
        ovlp_3c1e = ovlp_3c1e.reshape(-1, nao*nao, naux)
        
        # Combine j2c, ovlp_2c1e, and ovlp_3c1e
        j3c = numpy.zeros((chunk_size, naux, nao, nao), dtype=complex)
        for ik in range(local_kpairs_num):
            # find the corresponding q-point
            kij_idx = kij_idx_local[ik]
            q_idx = uniq_inverse[kij_idx]
            iq_red = numpy.argwhere(q_ir_list == q_index[q_idx])[0][0]
            j2c_local = j2c[iq_red] if q_conj_list[q_idx] == 0 else j2c[iq_red].conj()
            ovlp_2c1e_local = ovlp_2c1e_aux[iq_red] if q_conj_list[q_idx] == 0 else ovlp_2c1e_aux[iq_red].conj()
      
            # df_coef = (S^-1) * ovlp_3c1e: (naux, nao*nao)
            df_coef = LA.solve(ovlp_2c1e_local, ovlp_3c1e[ik].T)
            # j3c_buffer: (naux_effective, nao*nao).
            j3c[ik] = numpy.dot(j2c_local, df_coef).reshape(-1, nao, nao)
    
        VQ = h5py.File(basename + "/VQ_{}.h5".format(c0), 'w')
        VQ["{}".format(c0)] = j3c.view(float)
        VQ.close()
        c0 += chunk_size
        ovlp_3c1e[:] = 0.0
  
    data = h5py.File(filename, "w")
    data["chunk_size"] = chunk_size
    #data["chunk_indices"] = numpy.array(chunk_indices)
    data["grid/conj_pairs_list"] = kij_conj
    data["grid/trans_pairs_list"] = kij_trans
    data["grid/kpair_irre_list"] = kpair_irre_list
    data["grid/kpair_idx"] = kptij_idx
    data["grid/num_kpair_stored"] = num_kpair_stored
    data["grid/k_mesh"] = mydf.kpts
    data["grid/k_mesh_scaled"] = cell.get_scaled_kpts(mydf.kpts)
                                                             
    data.close()
    print("Integrals have been computed and stored into {}".format(basename))
  
    if j2c_sqrt:
        return j2c, uniq_qs
    else:
        return 0

def make_j3c(mydf, cell, j2c_sqrt=True, exx=False):
    '''
    The inefficient incore version of make_j3c
    '''
    make_j2c = _make_j2c_gdf
    nao = cell.nao_nr()
    naux = mydf.auxcell.nao_nr()
    
    kmesh_scaled = cell.get_scaled_kpts(mydf.kpts)
    kptij_lst = make_kptij_lst(kmesh_scaled)
    nkptij = len(kptij_lst)
    kptis = kptij_lst[:,0]
    kptjs = kptij_lst[:,1]
    kpt_ji = kptjs - kptis
    # Fold kpt_ji back to [-0.5, 0.5] BZ notation
    kpt_ji = comm.fold_back_to_1stBZ(kpt_ji)
    # uniq_qs: unique q-points, # uniq_inverse: kptij -> q
    uniq_qs, uniq_index, uniq_inverse = unique(kpt_ji)
    q_ir_list, q_index, _, q_conj_list = comm.inversion_sym(uniq_qs)
    print("number of k-pairs: {}".format(len(kptij_lst)))
    print("number of uniq q-pionts: ", len(uniq_qs))
    print("number of uniq q-points w/ the inversion symmetry: ", len(q_ir_list))
    
    kptij_lst = cell.get_abs_kpts(kptij_lst)
    uniq_qs = cell.get_abs_kpts(uniq_qs)
  
    # j2c: [q_ir_list][naux, naux]
    j2c = make_j2c(mydf, cell, mydf.auxcell, uniq_qs[q_ir_list], exx)
  
    # ovlp_2c1e: [q_ir_list][naux, naux]
    ovlp_2c1e_aux = mydf.auxcell.pbc_intor('int1e_ovlp', hermi=1, kpts=uniq_qs[q_ir_list])
  
    # ints_3c1e: [kptij_lst][nao*nao, naux]
    ints_3c1e = incore.aux_e2(cell, mydf.auxcell, intor='int3c1e', kptij_lst=kptij_lst)
    if nkptij == 1:
        ints_3c1e = ints_3c1e.reshape((1,)+ints_3c1e.shape)
  
    # Combine j2c, ints_2c1e, and ints_3c1e
  
    # j2c_sqrt: (naux_effective, naux). naux_effective <= naux due to linear dependency
    j2ctags = []
    for iq in range(len(q_ir_list)):
        j2c[iq], tag = compute_j2c_sqrt(iq, j2c[iq], mydf.linear_dep_threshold)
        j2ctags.append(tag)
  
    j3c = numpy.zeros((nkptij, naux, nao, nao), dtype=complex)
    for uniq_q_id in range(len(uniq_qs)):
        adapted_ji_idx = numpy.where(uniq_inverse == uniq_q_id)[0]
        print("uniq_kptij_id = {}".format(uniq_q_id))
        print("adapted_ji_idx = {}".format(adapted_ji_idx))
    
        iq_red = numpy.argwhere(q_ir_list == q_index[uniq_q_id])[0][0]
        j2c_local = j2c[iq_red] if q_conj_list[uniq_q_id] == 0 else j2c[iq_red].conj()
        ovlp_2c1e_local = ovlp_2c1e_aux[iq_red] if q_conj_list[uniq_q_id] == 0 else ovlp_2c1e_aux[iq_red].conj()
    
        for k, ji in enumerate(adapted_ji_idx):
            # ovlp_3c1e: [nao*nao, naux]
            ovlp_3c1e = ints_3c1e[ji]
            # df_coef = (S^-1) * ovlp_3c1e: (naux, nao*nao)
            df_coef = LA.solve(ovlp_2c1e_local, ovlp_3c1e.T)
            # j3c_buffer: (naux_effective, nao*nao).
            j3c_buffer = numpy.dot(j2c_local, df_coef)
            j3c[ji,:j3c_buffer.shape[0]] = j3c_buffer.reshape(-1, nao, nao)
  
    if j2c_sqrt:
        return j3c, kptij_lst, j2c, uniq_qs
    else:
        return j3c, kptij_lst


def check_eri(j3c1, j3c2, kptij_lst):
    nkptij = len(kptij_lst)                                  
    kptis = kptij_lst[:,0]
    kptjs = kptij_lst[:,1]
    kpt_ji = kptjs - kptis
    # uniq_qs: unique q-points, # uniq_inverse: kptij -> q
    uniq_qs, uniq_index, uniq_inverse = unique(kpt_ji)
  
    for uniq_kptji_id in range(len(uniq_qs)):
        print("qid = {}".format(uniq_kptji_id))
        adapted_ji_idx = numpy.where(uniq_inverse == uniq_kptji_id)[0]
        for k, ji in enumerate(adapted_ji_idx):
            eri1 = lib.einsum('Lpq,Lsr->pqrs', j3c1[ji], j3c1[ji].conj())
            eri2 = lib.einsum('Lpq,Lsr->pqrs', j3c2[ji], j3c2[ji].conj())
            diff = eri1 - eri2
            print(numpy.max(numpy.abs(diff)))
            if uniq_kptji_id == 2 and k == 0:
                print("eri_gdf_S:")
                print(eri1)
                print("eri_Coulomb")
                print(eri2)
                print("-----------")

def transformation_PW_to_auxbasis(mydf, cell, j2c_sqrt, qs):
    from functools import reduce
    nao, NQ = cell.nao_nr(), mydf.auxcell.nao_nr() 
  
    kmesh_scaled = cell.get_scaled_kpts(mydf.kpts)
    qs_scaled = cell.get_scaled_kpts(qs)
    ir_list, *_ = comm.inversion_sym(kmesh_scaled)
    q_ir_list, q_index, _, q_conj_list = comm.inversion_sym(qs_scaled)
    kmesh_reduced = mydf.kpts[ir_list]
    
    print("number of uniq q-pionts: ", len(qs))
    print("size of j2c_sqrt: {}".format(len(j2c_sqrt)))
    print("number of uniq q-points w/ the inversion symmetry: ", len(q_ir_list))
    
    # ints_2c1e_aux
    ints_2c1e_aux = mydf.auxcell.pbc_intor('int1e_ovlp', hermi=1, kpts=kmesh_reduced)
    
    AqQ = numpy.zeros((len(kmesh_reduced), NQ), dtype=complex)
    q_reduced = numpy.zeros((len(kmesh_reduced),3), dtype=float)
    q_scaled_reduced = numpy.zeros((len(kmesh_reduced),3), dtype=float)
    # q from k_mesh_reduced
    for ik, q in enumerate(kmesh_reduced):
        q_scaled = cell.get_scaled_kpts(q)
      
        # Test: Make sure q is in [-0.5, 0.5] 
        q_scaled = [comm.wrap_1stBZ(l) for l in q_scaled]
        q = cell.get_abs_kpts(q_scaled)
      
        # Chi_G0: (NQ)
        Chi_G0 = ft_ao.ft_ao(mydf.auxcell, Gv=numpy.zeros((1,3)), kpt=q)[0]
        S_aux_inv = LA.inv(ints_2c1e_aux[ik])
        uniq_q_id = -1
        # q from qs
        for iq2, q2_scaled in enumerate(qs_scaled):
            # q = [0.0, 1.0]
            q2_scaled = [comm.wrap_k(l) for l in q2_scaled]
            # q = [-0.5, 0.5]
            q2_scaled = [comm.wrap_1stBZ(l) for l in q2_scaled]
            if numpy.allclose(q_scaled, q2_scaled):
                iq2_red = numpy.argwhere(q_ir_list == q_index[iq2])[0][0]
                U = j2c_sqrt[iq2_red] if q_conj_list[iq2] == 0 else j2c_sqrt[iq2_red].conj()
                uniq_q_id = iq2
                break
        if U is None:
            raise ValueError("Error: No matching j2c_sqrt is found!")
        # A^q_Q = (U)^q * (S^{-1})^{q} * Chi_G0*
        q_reduced[ik] = q
        q_scaled_reduced[ik] = q_scaled
        # U: (naux_effective, naux), S_aux_inv: (naux, naux), Chi_G0: (naux)
        AqQ[ik, :U.shape[0]] = reduce(numpy.dot, (U, S_aux_inv, Chi_G0.conj()))
        AqQ[ik] *= (numpy.linalg.norm(q) / numpy.sqrt(4*numpy.pi*cell.vol))
        U = None

    return AqQ, q_reduced, q_scaled_reduced



from functools import reduce
import numpy as np
import irbasis
import h5py
import os

'''
Fourier transform between imaginary time and Matsubara frequency using
the intermediate representation (IR)
'''


class IR_factory(object):
    def __init__(self, beta, ir_file=None):

        if ir_file is None:
            raise ValueError(
                "{} is not an acceptable IR-grid file.".format(ir_file)
                + " Provide a valid hdf5 IR-grid file."
            )

        self.beta = beta
        self.ir_file = ir_file
        self.tau_mesh, self.wsample, self.Ttc, self.Tcn, \
            self.Tnc, self.Tct = read_IR_matrices(
                os.path.abspath(ir_file), self.beta
            )
        self.nts = self.tau_mesh.shape[0]
        self.nw = self.wsample.shape[0]

    def update(self, beta=None, ir_file=None):
        """Update IR grid information in run-time.
        """
        if ir_file is not None:
            self.ir_file = ir_file
        if beta is not None:
            self.beta = beta
        self.tau_mesh, self.wsample, self.Ttc, self.Tcn, \
            self.Tnc, self.Tct = read_IR_matrices(self.ir_file, self.beta)
        self.nts = self.tau_mesh.shape[0]
        self.nw = self.wsample.shape[0]

    def tau_to_w(self, X_t):
        """Transform `X_t` from tau to imaginary frequency representation.
        """
        X_w = np.zeros((self.nw,) + X_t.shape[1:], dtype=complex)
        original_shape = X_w.shape

        X_w, X_t = X_w.reshape(self.nw, -1), X_t.reshape(self.nts, -1)
        X_w = reduce(np.dot, (self.Tnc, self.Tct, X_t[1:-1]))
        X_w = X_w.reshape(original_shape)
        return X_w

    def w_to_tau(self, X_w, debug=False):
        """Transform `X_w` from imaginary frequency to tau representation.
        """
        X_t = np.zeros((self.nts,) + X_w.shape[1:], dtype=complex)
        original_shape = X_t.shape

        X_w, X_t = X_w.reshape(self.nw, -1), X_t.reshape(self.nts, -1)
        X_t = reduce(np.dot, (self.Ttc, self.Tcn, X_w))
        X_t = X_t.reshape(original_shape)
        if debug:
            # Check the imaginary parts
            print(
                "The largest imaginary parts in X_t is {}. Please double \
                check whether this is consistent to your expectation!".format(
                    np.max(np.abs(X_t.imag))
                )
            )
        return X_t

    def tauf_to_wb(self, X_t):
        """Transform quantity from fermionic tau-grid to bosonic
        frequency grid.
        E.g., P0(tau) -> P0(i Omega).
        """

        _, wsample_bose, Ttc_b, _, Tnc_b, Tct_b = read_IR_matrices(
                os.path.abspath(self.ir_file), self.beta, ptype='bose'
            )
        fir = h5py.File(self.ir_file, 'r')
        Ttc_other_tmp = fir['fermi/other_uxl'][()]
        nx_b = fir['bose/nx'][()]
        fir.close()
        Ttc_other = Ttc_b @ Tct_b @ Ttc_other_tmp
        Ttc_other *= np.sqrt(2.0 / self.beta)
        Tnt_bf = Tnc_b @ Tct_b @ Ttc_other[1:1 + nx_b, :] @ self.Tct

        nw = wsample_bose.shape[0]
        X_wb = np.zeros((nw,)+X_t.shape[1:], dtype=complex)
        original_shape = X_wb.shape

        X_wb, X_t = X_wb.reshape(nw, -1), X_t.reshape(self.nts, -1)
        X_wb = reduce(np.dot, (Tnt_bf, X_t[1:-1]))
        X_wb = X_wb.reshape(original_shape)

        return X_wb

    def wb_to_tauf(self, X_wb):
        """Transform quantity from bosonic frequency grid to fermionic
        tau-grid, e.g., P0(i Omega) -> P0(tau)
        """
        _, wsample_bose, _, Tcn_b, _, _ = read_IR_matrices(
                os.path.abspath(self.ir_file), self.beta, ptype='bose'
            )
        nw_b = len(wsample_bose)
        fir = h5py.File(self.ir_file, 'r')
        Ttc_b_other_tmp = fir['bose/other_uxl'][()]
        fir.close()
        Ttc_b_other = self.Ttc @ self.Tct @ Ttc_b_other_tmp
        Ttc_b_other *= np.sqrt(2.0 / self.beta)
        Ttn_fb = Ttc_b_other @ Tcn_b

        X_t = np.zeros((self.nts,) + X_wb.shape[1:], dtype=complex)
        original_shape = X_t.shape

        X_w, X_t = X_wb.reshape(nw_b, -1), X_t.reshape(self.nts, -1)
        X_t = reduce(np.dot, (Ttn_fb, X_w))
        X_t = X_t.reshape(original_shape)

        return X_t

    # TODO Specify the version of irbasis.
    def tau_to_w_other(self, X_t, wsample):
        """Use IR basis python package to intrinsically transform other type of
        quantities from tau to imaginary frequency basis.
        XXX: What are these "other" quantities?
        """
        nw = wsample.shape[0]
        X_w = np.zeros((nw,)+X_t.shape[1:], dtype=complex)
        original_shape = X_w.shape

        ir_factory = irbasis.load("F", float(self.ir_file))
        tnc = ir_factory.compute_unl(wsample)
        tnc *= np.sqrt(self.beta)

        X_w, X_t = X_w.reshape(nw, -1), X_t.reshape(self.nts, -1)
        X_w = reduce(np.dot, (tnc, self.Tct, X_t[1:-1]))
        X_w = X_w.reshape(original_shape)
        return X_w


def read_IR_matrices(ir_path, beta, ptype='fermi'):
    ir = h5py.File(ir_path, 'r')
    wsample = ir[ptype + "/wsample"][()]
    xsample = ir[ptype + "/xsample"][()]

    Ttc_minus1 = ir[ptype + "/ux1l_minus"][()]
    Ttc_tmp = ir[ptype + "/uxl"][()]
    Ttc_1 = ir[ptype + "/ux1l"][()]
    Ttc = np.zeros((Ttc_tmp.shape[0]+2, Ttc_tmp.shape[1]))
    Ttc[0], Ttc[1:-1], Ttc[-1] = Ttc_minus1, Ttc_tmp, Ttc_1
    Tnc_re = ir[ptype + "/uwl_re"][()]
    Tnc_im = ir[ptype + "/uwl_im"][()]
    Tnc = Tnc_re + 1j*Tnc_im
    ir.close()

    if ptype == 'fermi':
        zeta = 1
    else:
        zeta = 0
    wsample = (2*wsample + zeta) * np.pi / beta
    tau_mesh = np.zeros(xsample.shape[0]+2)
    tau_mesh[0], tau_mesh[1:-1], tau_mesh[-1] = 0, (xsample+1)*beta/2.0, beta

    Ttc *= np.sqrt(2.0/beta)
    Tnc *= np.sqrt(beta)
    Tct = np.linalg.inv(Ttc[1:-1])
    Tcn = np.linalg.inv(Tnc)

    return tau_mesh, wsample, Ttc, Tcn, Tnc, Tct



from functools import reduce
import numpy as np
import scipy.linalg as LA

"""
Orthogonalization utilities
"""


def canonical_matrices(S, thr=1e-7, type='f'):
    '''Löwdin's canonical orthogonalization'''
    # Form vectors for normalized overlap matrix
    Sval, Svec = LA.eigh(S)
    X = Svec[:, Sval >= thr] / np.sqrt(Sval[Sval >= thr])
    if type == 'f':
        return X
    elif type == 'g':
        X = LA.pinv(X)
        return X.T.conj()
    else:
        raise ValueError(
            "Invalid transofrmation type. Only 'f'/'g' type for \
            Fock/Green's function only."
        )


def canonical_orth(H, S, thr=1e-7, type='f'):
    '''Löwdin's canonical orthogonalization'''
    # if type != 'f':
    #     raise ValueError(
    #         "Invalid transformation type. "
    #         "Only 'f' type for Fock is supported currently."
    #     )
    print("Canonical orthogonalization with threshold = {}.".format(thr))
    ns = S.shape[0]
    nk = S.shape[1]
    nao = S.shape[2]
    original_shape = H.shape
    H = H.reshape(-1, ns, nk, nao, nao)
    H_orth = np.zeros(H.shape, dtype=H.dtype)
    for s in range(ns):
        for ik in range(nk):
            cond = np.linalg.cond(S[s, ik])
            if cond > 1e7:
                print(
                    "Warning: Condition number = {} is larger than"
                    " 1e7.".format(cond)
                )
            X = canonical_matrices(S[s, ik], thr, type)
            # nbands <= nao due to linear dependency
            nbands = X.shape[1]
            for d in range(H.shape[0]):
                H_orth[d, s, ik, :nbands, :nbands] = reduce(
                    np.dot, (X.T.conj(), H[d, s, ik], X)
                )
    H_orth = H_orth.reshape(original_shape)

    return H_orth


def sao_orth(X, S, type=None):
    """Symmetrized AO basis.
    """
    if not(type == 'g' or type == 'f'):
        raise ValueError(
            "Valid transformation types are 'g' for density, 'f' for Fock"
        )
    ns = S.shape[0]
    nk = S.shape[1]
    nao = S.shape[2]
    original_shape = X.shape
    X = X.reshape(-1, ns, nk, nao, nao)
    X_orth = np.zeros(X.shape, dtype=X.dtype)
    S12 = np.zeros(S.shape, dtype=S.dtype)
    for s in range(ns):
        for ik in range(nk):
            cond = np.linalg.cond(S[s, ik])
            if cond > 1e7:
                print(
                    "Warning: Condition number, {}, is larger than 1e7. "
                    "Possible numerical instability could appear. \n"
                    "Consider to use the canonical orthogonalization "
                    "instead.".format(cond)
                )
            if type == 'g':
                S12[s, ik] = LA.sqrtm(S[s, ik])
            elif type == 'f':
                S12[s, ik] = np.linalg.inv(LA.sqrtm(S[s, ik]))

    for d in range(X.shape[0]):
        for s in range(ns):
            for ik in range(nk):
                X_orth[d, s, ik] = reduce(
                    np.dot, (S12[s, ik], X[d, s, ik], S12[s, ik])
                )
    X_orth = X_orth.reshape(original_shape)
    return X_orth


def SAO_matrices(S):
    nk = S.shape[0]
    X_k = []
    X_inv_k = []

    for ik in range(nk):
        Sk = S[ik]
        x_pinv = LA.sqrtm(Sk)
        x = LA.inv(x_pinv)

        # store transformation basis for current k-point
        xx_inv = x_pinv.conj().T
        xx = x.conj().T
        X_inv_k.append(xx_inv.copy())
        X_k.append(xx.copy())

        # check that direct and inverse symmetries are consistent
        assert (np.allclose(np.dot(x, x_pinv), np.eye(x.shape[0])))
        assert (np.allclose(np.dot(X_inv_k[ik].conj().T, X_inv_k[ik]), S[ik]))

    return X_inv_k, X_k


def transform(Z, X, X_inv):
    '''
    Transform Z into X basis
    Z_X = X^* Z X
    :param Z: Object to be transformed
    :param X: Transformation matrix
    :param X_inv: Inverse transformation matrix
    :return: Z in new basis
    '''
    Z_X = np.zeros(Z.shape, dtype=complex)
    for ik in range(Z.shape[0]):
        Z_X[ik] = transform_per_k(Z[ik, :], X[ik], X_inv[ik])

    return Z_X


def transform_per_k(Z, X, X_inv):
    '''
    Transform Z into X basis
    Z_X = X^* Z X
    :param Z: Object to be transformed
    :param X: Transformation matrix
    :param X_inv: Inverse transformation matrix
    :return: Z in new basis
    '''
    # Z_X = np.einsum('ij,jk...,kl->il...', X, Z, X.T.conj())
    Z_X = np.dot(X.conj().T, np.dot(Z, X))

    Z_restore = X_inv.conj().T @ Z_X @ X_inv
    if not np.allclose(Z, Z_restore):
        error = "Orthogonal transformation failed. "\
            "Max difference between origin and restored quantity "\
            "is {}".format(np.max(Z - Z_restore))
        raise RuntimeError(error)
    return Z_X



import numpy as np

"""
Fourier transform between real and reciprocal space
"""


def construct_rmesh(nkx, nky, nkz):
    """Generat a real-space rmesh.
    rmesh is constructed in the units of lattice translation vectors.
    e.g.:   if nk = 6
            L = 2
            left = 1
            rx = -2, -1, 0, 1, 2, 3
    """
    Lx, Ly, Lz = (nkx - 1) // 2, (nky - 1) // 2, (nkz - 1) // 2
    leftx, lefty, leftz = (nkx - 1) % 2, (nky - 1) % 2, (nkz - 1) % 2

    rx = np.linspace(-Lx, Lx+leftx, nkx, endpoint=True)
    ry = np.linspace(-Ly, Ly+lefty, nky, endpoint=True)
    rz = np.linspace(-Lz, Lz+leftz, nkz, endpoint=True)
    RX, RY, RZ = np.meshgrid(rx, ry, rz)
    rmesh = np.array([RX.flatten(), RY.flatten(), RZ.flatten()]).T

    return rmesh


def construct_symmetric_rmesh(nkx, nky, nkz):
    """Generate a rmesh that is symmetric along the coordinate axes.
    e.g.:   if nk = 6
            L = 2
            left = 1
            rx_symm = -3, -2, -1, 0, 1, 2, 3
    """
    Lx, Ly, Lz = (nkx - 1) // 2, (nky - 1) // 2, (nkz - 1) // 2
    leftx, lefty, leftz = (nkx - 1) % 2, (nky - 1) % 2, (nkz - 1) % 2

    rx = np.linspace(-Lx, Lx + leftx, nkx, endpoint=True)
    ry = np.linspace(-Ly, Ly + lefty, nky, endpoint=True)
    rz = np.linspace(-Lz, Lz + leftz, nkz, endpoint=True)
    RX, RY, RZ = np.meshgrid(rx, ry, rz)
    rmesh = np.array([RX.flatten(), RY.flatten(), RZ.flatten()]).T
    pm_rmesh = np.append(rmesh, -rmesh, axis=0)
    rmesh = np.unique(pm_rmesh, axis=0)

    return rmesh


def compute_fourier_coefficients(kmesh, rmesh):
    """Compute Fourier coefficients for direct and
    inverse Fourier transform

    Input   :   kmesh, rmesh
    Returns :   fkr, frk
                where fkr = coefficients from r to k,
                and frk = coefficients from k to r.
    """
    fkr = np.zeros((kmesh.shape[0], rmesh.shape[0]), dtype=complex)
    frk = np.zeros((rmesh.shape[0], kmesh.shape[0]), dtype=complex)
    for ik, k in enumerate(kmesh):
        for jr, r in enumerate(rmesh):
            dp = 2.j*np.pi*np.dot(k, r)
            # coefficients from r to k
            fkr[ik, jr] = np.exp(-dp)
            # coefficients from k to r
            frk[jr, ik] = np.exp(dp)

    return fkr, frk


def k_to_real(frk, obj_k, weights):
    """
    Perform Fourier transform from reciprocal to real space
    frk     - Fourier transform coefficients from reciprocal to real space
    obj_k   - reciprocal space object
    weights - Fourier coefficient degenerate weights

    return obj_i - inverse Fourier transform of obj_k
    """
    obj_i = np.einsum("k...,k,ki->i...", obj_k, weights, frk.conj().T)
    obj_i /= np.sum(weights)
    return obj_i


def real_to_k(fkr, obj_i):
    """
    Perform Fourier transform from reciprocal to real space
    fkr     - Fourier transform coefficients from real to reciprocal space
    obj_i   - real space object

    return obj_k - Fourier transform of obj_i
    """
    original_shape = obj_i.shape
    obj_k = np.dot(fkr.conj(), obj_i.reshape(original_shape[0], -1))
    obj_k = obj_k.reshape((obj_k.shape[0],) + original_shape[1:])
    return obj_k



from functools import reduce
import numpy as np
import scipy.linalg as LA

from . import spectral as spec
from . import orth as orth
from .ir import IR_factory
from . import dyson as dyson
from . import winter as winter
from . import analyt_cont as AC


class MB_post(object):
    """Many-body analysis class"""
    def __init__(
        self, fock, sigma=None, mu=None, gtau=None, S=None, kmesh=None,
        beta=None, ir_file=None
    ):
        """ Initialization """
        # Public instance variables
        self.sigma = None
        self.fock = None
        self.S = None
        self.kmesh = None

        # Private instance variables
        self._gtau = None
        self._S_inv_12 = None
        self._nts = None
        self._ns = None
        self._ink = None
        self._nao = None
        self._ir_list = None
        self._weight = None
        self.ir = None
        self._ir_file = None
        self._beta = None
        self._mu = None

        """Setup"""
        if fock.ndim == 4:
            # For UHF data
            self._ns = fock.shape[0]
        elif fock.ndim == 3:
            # For RHF data, reshape matrices are originally in the shape:
            #   self_energy / gtau  :   (ntau, nk, nao, nao)
            #   fock                :   (nk, nao, nao)
            # To make RHF/UHF analysis easier, we transform the dimensions to:
            #   self_energy / gtau  :   (ntau, ns=1, nk, nao, nao)
            #   fock                :   (ns=1, nk, nao, nao)
            self._ns = 1
            fock = fock.reshape((1,) + fock.shape)
            if gtau is not None:
                gtau = gtau.reshape((gtau.shape[0], 1) + gtau.shape[1:])
            if sigma is not None:
                sigma = sigma.reshape((sigma.shape[0], 1) + sigma.shape[1:])
            if S is not None:
                S = S.reshape((1,) + S.shape)
        else:
            raise ValueError(
                'Incorrect dimensions of self-energy or Fock. Acceptable \
                shapes are (nts, ns, nk, nao, nao) or (nts, nk, nao, nao) \
                for self energy and (ns, nk, nao, nao) or (nk, nao, nao) \
                for Fock matrix.'
            )

        if mu is None:
            print("Warning: Default chemical potential, mu = 0.0, is used.")
            self._mu = 0.0
        else:
            self._mu = mu

        if beta is None:
            print("Warning: Inverse temperature is set to the default value\
                1000 a.u.^{-1}.")
            self._beta = 1000
        else:
            self._beta = beta

        if ir_file is None:
            raise ValueError(
                "{} is not an acceptable IR-grid file.".format(ir_file)
                + " Provide a valid hdf5 IR-grid file."
            )
        else:
            self.ir_file = ir_file

        self._ink = fock.shape[1]
        self._nao = fock.shape[2]
        self._ir_list = np.arange(self._ink)
        self._weight = np.array([1 for i in range(self._ink)])

        self.fock = fock.copy()
        if sigma is not None:
            self.sigma = sigma.copy()
        if S is not None:
            self.S = S.copy()
        if kmesh is not None:
            self.kmesh = kmesh.copy()
        if gtau is not None:
            self.gtau = gtau.copy()

        print(self)

    @property
    def beta(self):
        return self._beta

    @beta.setter
    def beta(self, value):
        """
        Changing beta will automatically update self.ir for consistency
        """
        print("Updated beta = {}".format(value))
        self._beta = value
        if self.ir is None:
            self.ir = IR_factory(self.beta, self.ir_file)
        else:
            self.ir.update(self.beta, self.ir_file)

    @property
    def ir_file(self):
        return self._ir_file

    @ir_file.setter
    def ir_file(self, value):
        """Changing ir_file will automatically update both self._nts
        and self.ir for consistency.
        :param value: Path to the IR grid HDF5 file.
        :return:
        """
        print("Setting up IR grid for {}".format(value))
        self._ir_file = value
        if self.ir is None:
            self.ir = IR_factory(self.beta, self.ir_file)
        else:
            self.ir.update(self.beta, self.ir_file)
        self._nts = self.ir.nts

    @property
    def mu(self):
        return self._mu

    @mu.setter
    def mu(self, value):
        """Updating chemical potential implicitly implies updating the
        Green's function and density matrix.
        :param value:
        :return:
        """
        print("Updated mu = {}".format(value))
        self._mu = value
        if self._gtau is not None:  # or self.dm is not None:
            self.solve_dyson()

    @property
    def gtau(self):
        if self._gtau is None:
            self.solve_dyson()
        return self._gtau

    @gtau.setter
    def gtau(self, G):
        """
        Updating gtau will implicitly update density matrix (dm).
        :param G:
        :return:
        """
        self._gtau = G

    @property
    def dm(self):
        return -1.0 * self.gtau[-1]

    def __str__(self):
        return "######### MBPT analysis class #########\n" \
            "nts  = {} \n" \
            "ns   = {}\n" \
            "nk   = {}\n" \
            "nao  = {}\n" \
            "mu   = {}\n" \
            "beta = {}\n" \
            "#######################################".format(
                self._nts, self._ns, self._ink, self._nao, self.mu, self.beta
            )

    def solve_dyson(self):
        """
        Compute Green's function through Dyson's equation and update self.gtau
        and self.dm.
        :return:
        """
        self.gtau = dyson.solve_dyson(
            self.fock, self.S, self.sigma, self.mu, self.ir
        )

    def eigh(self, F, S, thr=1e-7):
        return LA.eigh(F, S)
    # FIXME c here is differ with the c from eigh() by a phase factor.
    # Fix it or leave it like this?

    def eigh_canonical(self, F, S, thr=1e-7):
        # S: m*m, x =: m*n, xFx: n*n, c: n*n, e: n, xc: m*n
        x = orth.canonical_matrices(S, thr, 'f')
        xFx = reduce(np.dot, (x.T.conj(), F, x))
        e, c = LA.eigh(xFx)
        c = np.dot(x, c)

        return e, c

    def get_mo(self, canonical=False, thr=1e-7):
        """
        Compute molecular orbital energy by solving FC=SCE
        :return:
        """
        if not canonical:
            eigh = self.eigh
        else:
            eigh = self.eigh_canonical
        mo_energy, mo_coeff = spec.compute_mo(self.fock, self.S, eigh, thr)
        return mo_energy, mo_coeff

    def get_no(self):
        """
        Compute natural orbitals by diagonalizing density matrix
        :return:
        """
        occ, no_coeff = spec.compute_no(self.dm, self.S)
        return occ, no_coeff

    def mulliken_analysis(self, orbitals=None):
        if orbitals is None:
            orbitals = np.arange(self._nao)
        occupations = np.zeros((self._ns, orbitals.shape[0]), dtype=complex)
        if self.S is not None:
            occupations = np.einsum(
                'k,skij,skji->si', self._weight, self.dm, self.S
            )
        else:
            occupations = np.einsum('k,skii->si', self._weight, self.dm)
        num_k = len(self._weight)
        occupations /= num_k

        # Check imaginary part
        imag = np.max(np.abs(occupations.imag))
        print("The maximum of imaginary part is ", imag)

        return occupations.real

    def wannier_interpolation(self, kpts_inter, hermi=False, debug=False):
        """
        Wannier interpolation
        :param kpts_int: Scaled k-points for the target k grid
        :return:
        """
        if self.kmesh is None:
            raise ValueError(
                "kmesh of input data is unknown. Please provide it."
            )
        Gtk_int, Sigma_tk_int, tau_mesh, Fk_int, Sk_int = winter.interpolate_G(
            self.fock, self.sigma, self.mu, self.S,
            self.kmesh, kpts_inter, self.ir, hermi=hermi, debug=debug
        )
        return Gtk_int, Sigma_tk_int, tau_mesh, Fk_int, Sk_int

    def AC_maxent(
        self, error=5e-3, maxent_exe='maxent', params='green.param',
        outdir='Maxent', gtau_orth=None
    ):
        """
        Analytical continuation using Maxent
        :param error:
        :param maxent_exe:
        :param params:
        :param outdir:
        :param gtau:
        :return:
        """
        if gtau_orth is None:
            gtau_orth = orth.sao_orth(
                self.gtau, self.S, type='g'
            ) if self.S is not None else self.gtau
            gtau_inp = np.einsum("...ii->...i", gtau_orth)
        else:
            gtau_inp = gtau_orth
        tau_mesh = self.ir.tau_mesh

        AC.maxent_run(gtau_inp, tau_mesh, error, params, maxent_exe, outdir)

    def AC_nevanlinna(
        self, outdir="Nevanlinna", gtau_orth=None,
        n_real=10001, w_min=-10., w_max=10., eta=0.01
    ):
        """
        Analytical continuation using Nevanlinna interpolation
        :param nevan_exe:
        :param outdir:
        :return:
        """
        if gtau_orth is None:
            gtau_orth = orth.sao_orth(
                self.gtau, self.S, type='g'
            ) if self.S is not None else self.gtau
            gtau_orth = np.einsum("...ii->...i", gtau_orth)
        nw = self.ir.wsample.shape[0]
        Gw_inp = self.ir.tau_to_w(gtau_orth)[nw//2:]

        wsample = self.ir.wsample[nw//2:]
        freqs, A_w = AC.nevan_run(
            Gw_inp, wsample, outdir=outdir, n_real=n_real,
            w_min=w_min, w_max=w_max, eta=eta, spectral=True
        )

        return freqs, A_w


def minus_k_to_k_TRsym(X):
    nso = X.shape[-1]
    nao = nso // 2
    Y = np.zeros(X.shape, dtype=X.dtype)
    Y[:nao, :nao] = X[nao:, nao:].conj()
    Y[nao:, nao:] = X[:nao, :nao].conj()
    Y[:nao, nao:] = -1.0 * X[nao:, :nao].conj()
    Y[nao:, :nao] = Y[:nao, nao:].conj().transpose()
    return Y


def to_full_bz_TRsym(X, conj_list, ir_list, bz_index, k_ind):
    index_list = np.zeros(bz_index.shape, dtype=int)
    for i, irn in enumerate(ir_list):
        index_list[irn] = i
    old_shape = X.shape
    new_shape = np.copy(old_shape)
    new_shape[k_ind] = conj_list.shape[0]
    Y = np.zeros(new_shape, dtype=X.dtype)
    for ik, kk in enumerate(bz_index):
        k = index_list[kk]
        Y = Y.reshape((-1,) + Y.shape[k_ind:])
        X = X.reshape((-1,) + X.shape[k_ind:])
        for i in range(Y.shape[0]):
            Y[i, ik] = minus_k_to_k_TRsym(
                X[i, k]
            ) if conj_list[ik] else X[i, k]
        Y = Y.reshape(new_shape)
        X = X.reshape(old_shape)

    return Y


def to_full_bz(X, conj_list, ir_list, bz_index, k_ind):
    """Transform input quantity from irreducible number of k-points
    to the entire Brillouin Zone.
    """
    index_list = np.zeros(bz_index.shape, dtype=int)
    for i, irn in enumerate(ir_list):
        index_list[irn] = i
    old_shape = X.shape
    new_shape = np.copy(old_shape)
    new_shape[k_ind] = conj_list.shape[0]
    Y = np.zeros(new_shape, dtype=X.dtype)
    for ik, kk in enumerate(bz_index):
        k = index_list[kk]
        if k_ind == 0:
            Y[ik, ::] = X[k, ::].conj() if conj_list[ik] else X[k, ::]
        elif k_ind == 1:
            Y[:, ik, ::] = X[:, k, ::].conj() \
                if conj_list[ik] else X[:, k, ::]
        elif k_ind == 2:
            Y[:, :, ik, ::] = X[:, :, k, ::].conj() \
                if conj_list[ik] else X[:, :, k, ::]
    return Y


def initialize_MB_post(sim_path=None, input_path=None, ir_file=None):
    import h5py
    f = h5py.File(sim_path, 'r')
    it = f["iter"][()]
    Sr = f["S-k"][()].view(complex)
    Sr = Sr.reshape(Sr.shape[:-1])
    Fr = f["iter"+str(it)+"/Fock-k"][()].view(complex)
    Fr = Fr.reshape(Fr.shape[:-1])
    Sigmar = f["iter"+str(it)+"/Selfenergy/data"][()].view(complex)
    Sigmar = Sigmar.reshape(Sigmar.shape[:-1])
    Gr = f["iter"+str(it)+"/G_tau/data"][()].view(complex)
    Gr = Gr.reshape(Gr.shape[:-1])
    tau_mesh = f["iter"+str(it)+"/G_tau/mesh"][()]
    beta = tau_mesh[-1]
    mu = f["iter"+str(it)+"/mu"][()]
    f.close()

    f = h5py.File(input_path, 'r')
    ir_list = f["/grid/ir_list"][()]
    index = f["/grid/index"][()]
    conj_list = f["grid/conj_list"][()]
    f.close()

    """ All k-dependent matrices should lie on a full Monkhorst-Pack grid. """
    F = to_full_bz(Fr, conj_list, ir_list, index, 1)
    S = to_full_bz(Sr, conj_list, ir_list, index, 1)
    Sigma = to_full_bz(Sigmar, conj_list, ir_list, index, 2)
    G = to_full_bz(Gr, conj_list, ir_list, index, 2)

    """ Results from correlated methods """

    # Standard way to initialize
    return MB_post(
        fock=F, sigma=Sigma, mu=mu, gtau=G, S=S, beta=beta, ir_file=ir_file
    )



import numpy as np


def solve_dyson(fock, S=None, sigma=None, mu=0, ir=None):
    """
    Compute Green's function through Dyson's equation.
    :return:
    """
    nts, nw = ir.nts, ir.nw
    ns, nk, nao = fock.shape[0], fock.shape[1], fock.shape[2]
    if S is None:
        S = np.array([[np.eye(nao)]*nk]*ns)

    gtau = np.zeros((nts, ns, nk, nao, nao), dtype=complex)
    Gw = np.zeros((nw, nao, nao), dtype=complex)
    for s in range(ns):
        for k in range(nk):
            if sigma is not None:
                sigma_w = ir.tau_to_w(sigma[:, s, k])
            for n in range(nw):
                if sigma is None:
                    tmp = (1j * ir.wsample[n] + mu) * S[s, k] - fock[s, k]
                else:
                    tmp = (1j * ir.wsample[n] + mu) * S[s, k] \
                        - fock[s, k] - sigma_w[n]
                Gw[n] = np.linalg.inv(tmp)
            gtau[:, s, k] = ir.w_to_tau(Gw)

    return gtau
